/*!
 * Copyright (c) <2018> <Andreas Niskanen, University of Helsinki>
 * 
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * 
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "ArguFramework.h"

#include <iostream>
#include <algorithm>


AF::AF() : n_args(0), count(0) {}

/*!
 * Adds a new argument to the AF instance.
 */
void AF::addArgument(std::string arg)
{
    args.push_back(++n_args);
    argToInt[arg] = n_args;
    intToArg[n_args] = arg;
}

/*!
 * Adds a new attack to the AF instance.
 */
void AF::addAttack(std::pair<std::string,std::string> att)
{
    atts.push_back(std::make_pair(argToInt[att.first], argToInt[att.second]));
    attackers[argToInt[att.second]].push_back(argToInt[att.first]);
    attacked[argToInt[att.first]].push_back(argToInt[att.second]);
    att_exists[std::make_pair(argToInt[att.first], argToInt[att.second])] = true;
}

void AF::initialize()
{
    for (int i = 0; i < args.size(); i++) {
        argToVar[args[i]] = ++count;
        varToArg[count] = args[i];
    }
    for (int i = 0; i < args.size(); i++) {
        defendVar[args[i]] = ++count;
    }
}

bool AF::is_cf(std::vector<int> & extension)
{
    for (int i = 0; i < extension.size(); i++) {
        for (int j = 0; j < extension.size(); j++) {
            if (att_exists[std::make_pair(extension[i],extension[j])]) {
                return false;
            }
        }
    }
    return true;
}

bool AF::is_adm(std::vector<int> & extension)
{
    if (!is_cf(extension)) return false;
    for (int i = 0; i < extension.size(); i++) {
        for (int j = 0; j < args.size(); j++) {
            if (att_exists[std::make_pair(args[j],extension[i])]) {
                bool flag = false;
                for (int k = 0; k < extension.size(); k++) {
                    if (att_exists[std::make_pair(extension[k],args[j])]) {
                        flag = true;
                        break;
                    }
                }
                if (!flag) return false;
            }
        }
    }
    return true;
}

bool AF::is_com(std::vector<int> & extension)
{
    if (!is_adm(extension)) return false;
    for (int i = 0; i < args.size(); i++) {
        if (std::find(extension.begin(), extension.end(), args[i]) == extension.end()) {
            if (attackers[args[i]].size() == 0) return false;
            for (int j = 0; j < attackers[args[i]].size(); j++) {
                for (int k = 0; k < extension.size(); k++) {
                    if (att_exists[std::make_pair(extension[k], attackers[args[i]][j])]) {
                        return false;
                    }
                }
            }
        }
    }
    return true;
}

void AF::print()
{
    for (int i = 0; i < args.size(); i++) {
        std::cout << "arg(" << intToArg[args[i]] << ").\n";
    }
    for (int i = 0; i < atts.size(); i++) {
        std::cout << "att(" << intToArg[atts[i].first] << "," << intToArg[atts[i].second] << ").\n";
    }
}