/*!
 * Copyright (c) <2018> <Andreas Niskanen, University of Helsinki>
 * 
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * 
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef ARGU_FRAMEWORK_H
#define ARGU_FRAMEWORK_H

#include <vector>
#include <map>

class AF {
public:
    
AF();

/*!
 * Number of arguments and count for MaxSAT variables.
 */    
int n_args;
int count;

/*!
 * Arguments and attacks of the AF instance.
 */
std::vector<int> args;
std::vector<std::pair<int,int>> atts;

/*!
 * Maps an argument to an integer and the other way around.
 */
std::map<std::string,int> argToInt;
std::map<int,std::string> intToArg;

/*!
 * Maps an attack to true if the attack exists.
 */
std::map<std::pair<int,int>,bool> att_exists;

/*!
 * Maps an argument to its attackers.
 */
std::map<int,std::vector<int>> attackers;
std::map<int,std::vector<int>> attacked;

/*!
 * Maps an argument to its MaxSAT variable and the other way around.
 */
std::map<int,int> argToVar;
std::map<int,int> varToArg;

/*!
 * Other MaxSAT variables.
 */
std::map<int,int> defendVar;

/*!
 * Adds a new argument to the AF instance.
 */
void addArgument(std::string arg);

/*!
 * Adds a new attack to the AF instance.
 */
void addAttack(std::pair<std::string,std::string> att);

/*!
 * Constructs the variable mappings needed for generation of MaxSAT clauses.
 */
void initialize();

bool is_cf(std::vector<int> & extension);
bool is_adm(std::vector<int> & extension);
bool is_com(std::vector<int> & extension);

void print();

};

#endif
