/*!
 * Copyright (c) <2018> <Andreas Niskanen, University of Helsinki>
 * 
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * 
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "Instance.h"

#include <iostream>

void Instance::add_argument(std::string arg)
{
    arg_to_index[arg] = arguments.size()+1;
    index_to_arg[arguments.size()+1] = arg;
    arguments.push_back(arguments.size()+1);
}

void Instance::add_example(Example ex)
{
    example_id_to_index[ex.id] = examples.size();
    index_to_example_id[examples.size()] = ex.id;
    examples.push_back(ex);
}

void Instance::initialize(bool adm_in_abstraction)
{
	for (int i = 0; i < arguments.size(); i++) {
		for (int j = 0; j < arguments.size(); j++) {
			attack_var[std::make_pair(arguments[i], arguments[j])] = ++count;
			var_attack[count] = std::make_pair(arguments[i], arguments[j]);
		}
	}
	for (int i = 0; i < examples.size(); i++) {
		extension_var[examples[i].id] = ++count;
		var_extension[count] = examples[i].id;
	}
	for (int i = 0; i < examples.size(); i++) {
		if (examples[i].pos && (examples[i].sem == "com" || (!adm_in_abstraction && examples[i].sem == "prf"))) {
			for (int j = 0; j < arguments.size(); j++) {
				if (!examples[i].arg_in_ex[arguments[j]]) {
					for (int k = 0; k < arguments.size(); k++) {
						not_defended_var[std::make_pair(examples[i].id, std::make_pair(arguments[j], arguments[k]))] = ++count;
					}
				}
			}
		}
	}
	for (int i = 0; i < examples.size(); i++) {
		if (!examples[i].pos && (examples[i].sem == "adm" || examples[i].sem == "prf")) {
			for (int j = 0; j < arguments.size(); j++) {
				if (examples[i].arg_in_ex[arguments[j]]) {
					for (int k = 0; k < arguments.size(); k++) {
						if (!examples[i].arg_in_ex[arguments[k]]) {
							not_defended_var[std::make_pair(examples[i].id, std::make_pair(arguments[j], arguments[k]))] = ++count;
						}
					}
				}
			}
		}
	}
	for (int i = 0; i < examples.size(); i++) {
		if (!examples[i].pos && examples[i].sem == "com") {
			for (int j = 0; j < arguments.size(); j++) {
				for (int k = 0; k < arguments.size(); k++) {
					not_defended_var[std::make_pair(examples[i].id, std::make_pair(arguments[j], arguments[k]))] = ++count;
				}
			}
		}
	}
	for (int i = 0; i < examples.size(); i++) {
		if (!examples[i].pos && examples[i].sem == "com") {
			for (int j = 0; j < arguments.size(); j++) {
				if (!examples[i].arg_in_ex[arguments[j]]) {
					exists_not_defended_var[std::make_pair(examples[i].id, arguments[j])] = ++count;
				}
			}
		}
	}
	for (int i = 0; i < examples.size(); i++) {
		if (!examples[i].pos && examples[i].sem == "stb") {
			for (int j = 0; j < arguments.size(); j++) {
				if (!examples[i].arg_in_ex[arguments[j]]) {
					not_attacked_var[std::make_pair(examples[i].id, arguments[j])] = ++count;
				}
			}
		}
	}
	for (int i = 0; i < examples.size(); i++) {
		if (!examples[i].pos && examples[i].sem == "prf") {
			exists_superset_var[examples[i].id] = ++count;
			var_exists_superset[count] = examples[i].id;
			for (int j = 0; j < arguments.size(); j++) {
               	if (!examples[i].arg_in_ex[arguments[j]]) {
                   	arg_var[std::make_pair(examples[i].id,arguments[j])] = ++count;
               	}
           	}
           	for (int j = 0; j < arguments.size(); j++) {
               	for (int k = 0; k < arguments.size(); k++) {
                   	if (!examples[i].arg_in_ex[arguments[j]] && !examples[i].arg_in_ex[arguments[k]]) {
                       	attacked_var[std::make_pair(examples[i].id,std::make_pair(arguments[j], arguments[k]))] = ++count;
                   	}
               	}
           	}
           	for (int j = 0; j < arguments.size(); j++) {
               	for (int k = 0; k < arguments.size(); k++) {
                   	if (!examples[i].arg_in_ex[arguments[j]] && !examples[i].arg_in_ex[arguments[k]]) {
                       	attacks_var[std::make_pair(examples[i].id,std::make_pair(arguments[j], arguments[k]))] = ++count;
                   	}
               	}
           	}
		}
	}
}

void Instance::print()
{
	std::cout << "Arguments:\n";
	for (int i = 0; i < arguments.size(); i++) {
		std::cout << "arg " << index_to_arg[arguments[i]] << "\n";
	}
	std::cout << "Examples:\n";
	for (int i = 0; i < examples.size(); i++) {
		std::cout << examples[i].pos << " ex " << examples[i].id << "\n";
		std::cout << "weight " << examples[i].weight << "\n";
		std::cout << "sem " << examples[i].sem << "\n";
		std::cout << "args ";
		for (int j = 0; j < arguments.size(); j++) {
			if (examples[i].arg_in_ex[arguments[j]]) {
				std::cout << index_to_arg[arguments[j]] << " ";
			}
		}
		std::cout << "\n";
	}
}