/*!
 * Copyright (c) <2018> <Andreas Niskanen, University of Helsinki>
 * 
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * 
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "Synthesis.h"
#include "Verification.h"

#if defined(MAXSAT_LMHS)
#include "LMHSSolver.h"
typedef LMHSSolver MaxSAT_Solver;
#else
#error "No MaxSAT solver defined"
#endif

#if defined(SAT_MINISAT)
#include "MiniSATSolver.h"
typedef MiniSATSolver SAT_Solver;
#else
#error "No SAT solver defined"
#endif

#include <iostream>
#include <fstream>

namespace Synthesis {

using namespace std;

vector<vector<int>> cf_clauses(Instance & inst, Example & ex)
{
	vector<vector<int>> clauses;
	if (ex.pos) {
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
					if (ex.arg_in_ex[inst.arguments[j]]) {
						vector<int> clause;
						clause.push_back(-inst.extension_var[ex.id]);
						clause.push_back(-inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
						clauses.push_back(clause);
					}
				}
			}
		}
	}
	if (!ex.pos) {
		vector<int> clause;
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
					if (ex.arg_in_ex[inst.arguments[j]]) {
						clause.push_back(inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
					}
				}
			}
		}
		clause.push_back(inst.extension_var[ex.id]);
		clauses.push_back(clause);
	}
	return clauses;
}

vector<vector<int>> adm_clauses(Instance & inst, Example & ex)
{
	vector<vector<int>> clauses;
	if (ex.pos) {
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (ex.arg_in_ex[inst.arguments[i]]) {
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (ex.arg_in_ex[inst.arguments[j]]) {
					vector<int> clause;
					clause.push_back(-inst.extension_var[ex.id]);
					clause.push_back(-inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
					clauses.push_back(clause);
				}
			}
		}
	}
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (ex.arg_in_ex[inst.arguments[i]]) {
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (!ex.arg_in_ex[inst.arguments[j]]) {
					vector<int> clause;
					clause.push_back(-inst.extension_var[ex.id]);
					clause.push_back(-inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
					for (int k = 0; k < inst.arguments.size(); k++) {
						if (ex.arg_in_ex[inst.arguments[k]]) {
							clause.push_back(inst.attack_var[make_pair(inst.arguments[k], inst.arguments[j])]);
						}
					}
					clauses.push_back(clause);
				}
			}
		}
	}
	}
	if (!ex.pos) {
	vector<int> clause;
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (ex.arg_in_ex[inst.arguments[i]]) {
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (ex.arg_in_ex[inst.arguments[j]]) {
					clause.push_back(inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
				}
			}
		}
	}
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (ex.arg_in_ex[inst.arguments[i]]) {
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (!ex.arg_in_ex[inst.arguments[j]]) {
					clause.push_back(inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
				}
			}
		}
	}
	clause.push_back(inst.extension_var[ex.id]);
	clauses.push_back(clause);
	// here we define not defended var ???
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (ex.arg_in_ex[inst.arguments[i]]) {
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (!ex.arg_in_ex[inst.arguments[j]]) {
					vector<int> clause;
					clause.push_back(-inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
					clause.push_back(inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
					clauses.push_back(clause);
					for (int k = 0; k < inst.arguments.size(); k++) {
						if (ex.arg_in_ex[inst.arguments[k]]) {
							vector<int> clause;
							clause.push_back(-inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
							clause.push_back(-inst.attack_var[make_pair(inst.arguments[k], inst.arguments[j])]);
							clauses.push_back(clause);
						}
					}
				}
			}
		}
	}
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (ex.arg_in_ex[inst.arguments[i]]) {
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (!ex.arg_in_ex[inst.arguments[j]]) {
					vector<int> clause;
					clause.push_back(-inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
					for (int k = 0; k < inst.arguments.size(); k++) {
						if (ex.arg_in_ex[inst.arguments[k]]) {
							clause.push_back(inst.attack_var[make_pair(inst.arguments[k], inst.arguments[j])]);
						}
					}
					clause.push_back(inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
					clauses.push_back(clause);
				}
			}
		}
	}
	}
	return clauses;
}

vector<vector<int>> com_clauses(Instance & inst, Example & ex)
{
	vector<vector<int>> clauses;
	// positive examples
	if (ex.pos) {
		// conflict freeness
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
					if (ex.arg_in_ex[inst.arguments[j]]) {
						vector<int> clause;
						clause.push_back(-inst.extension_var[ex.id]);
						clause.push_back(-inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
						clauses.push_back(clause);
					}
				}
			}
		}
		// admissibility
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
					if (!ex.arg_in_ex[inst.arguments[j]]) {
						vector<int> clause;
						clause.push_back(-inst.extension_var[ex.id]);
						clause.push_back(-inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
						for (int k = 0; k < inst.arguments.size(); k++) {
							if (ex.arg_in_ex[inst.arguments[k]]) {
								clause.push_back(inst.attack_var[make_pair(inst.arguments[k], inst.arguments[j])]);
							}
						}
						clauses.push_back(clause);
					}
				}
			}
		}
		// completeness
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (!ex.arg_in_ex[inst.arguments[i]]) {
				vector<int> clause;
				clause.push_back(-inst.extension_var[ex.id]);
				for (int j = 0; j < inst.arguments.size(); j++) {
					clause.push_back(inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
				}
				clauses.push_back(clause);
			}
		}
		// define not defended
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (!ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
					vector<int> clause;
					clause.push_back(-inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
					clause.push_back(inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
					clauses.push_back(clause);
					for (int k = 0; k < inst.arguments.size(); k++) {
						if (ex.arg_in_ex[inst.arguments[k]]) {
							vector<int> clause;
							clause.push_back(-inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
							clause.push_back(-inst.attack_var[make_pair(inst.arguments[k], inst.arguments[j])]);
							clauses.push_back(clause);
						}
					}
				}
			}
		}
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (!ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
					vector<int> clause;
					clause.push_back(-inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
					for (int k = 0; k < inst.arguments.size(); k++) {
						if (ex.arg_in_ex[inst.arguments[k]]) {
							clause.push_back(inst.attack_var[make_pair(inst.arguments[k], inst.arguments[j])]);
						}
					}
					clause.push_back(inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
					clauses.push_back(clause);
				}
			}
		}
	}
	// negative example
	if (!ex.pos) {
		vector<int> clause;
		// conflict freeness
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
					if (ex.arg_in_ex[inst.arguments[j]]) {
						clause.push_back(inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
					}
				}
			}
		}
		// admissibility
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
					if (!ex.arg_in_ex[inst.arguments[j]]) {
						clause.push_back(inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
					}
				}
			}
		}
		// completeness
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (!ex.arg_in_ex[inst.arguments[i]]) {
				clause.push_back(-inst.exists_not_defended_var[make_pair(ex.id, inst.arguments[i])]);		
			}
		}
		clause.push_back(inst.extension_var[ex.id]);
		clauses.push_back(clause);
		// define not defended
		for (int i = 0; i < inst.arguments.size(); i++) {
			//if (ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
			//		if (!ex.arg_in_ex[inst.arguments[j]]) {
						vector<int> clause;
						clause.push_back(-inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
						clause.push_back(inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
						clauses.push_back(clause);
						for (int k = 0; k < inst.arguments.size(); k++) {
							if (ex.arg_in_ex[inst.arguments[k]]) {
								vector<int> clause;
								clause.push_back(-inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
								clause.push_back(-inst.attack_var[make_pair(inst.arguments[k], inst.arguments[j])]);
								clauses.push_back(clause);
							}
						}
			//		}
				}
			//}
		}
		for (int i = 0; i < inst.arguments.size(); i++) {
			//if (ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
			//		if (!ex.arg_in_ex[inst.arguments[j]]) {
						vector<int> clause;
						clause.push_back(-inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
						for (int k = 0; k < inst.arguments.size(); k++) {
							if (ex.arg_in_ex[inst.arguments[k]]) {
								clause.push_back(inst.attack_var[make_pair(inst.arguments[k], inst.arguments[j])]);
							}
						}
						clause.push_back(inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
						clauses.push_back(clause);
			//		}
				}
			//}
		}
		// define exists not defended
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (!ex.arg_in_ex[inst.arguments[i]]) {
				vector<int> clause;
				clause.push_back(-inst.exists_not_defended_var[make_pair(ex.id, inst.arguments[i])]);
				for (int j = 0; j < inst.arguments.size(); j++) {
					clause.push_back(inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
				}
				clauses.push_back(clause);
			}
		}
		for (int i = 0; i < inst.arguments.size(); i++) {
			if (!ex.arg_in_ex[inst.arguments[i]]) {
				for (int j = 0; j < inst.arguments.size(); j++) {
					vector<int> clause;
					clause.push_back(-inst.not_defended_var[make_pair(ex.id, make_pair(inst.arguments[i], inst.arguments[j]))]);
					clause.push_back(inst.exists_not_defended_var[make_pair(ex.id, inst.arguments[i])]);
					clauses.push_back(clause);
				}
			}
		}
	}
	return clauses;
}

vector<vector<int>> stb_clauses(Instance & inst, Example & ex)
{
	vector<vector<int>> clauses;
	if (ex.pos) {
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (ex.arg_in_ex[inst.arguments[i]]) {
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (ex.arg_in_ex[inst.arguments[j]]) {
					vector<int> clause;
					clause.push_back(-inst.extension_var[ex.id]);
					clause.push_back(-inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
					clauses.push_back(clause);
				}
			}
		}
	}
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (!ex.arg_in_ex[inst.arguments[i]]) {
			vector<int> clause;
			clause.push_back(-inst.extension_var[ex.id]);
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (ex.arg_in_ex[inst.arguments[j]]) {
					clause.push_back(inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
				}
			}
			clauses.push_back(clause);
		}
	}
	}
	if (!ex.pos) {
	vector<int> clause;
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (ex.arg_in_ex[inst.arguments[i]]) {
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (ex.arg_in_ex[inst.arguments[j]]) {
					clause.push_back(inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
				}
			}
		}
	}
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (!ex.arg_in_ex[inst.arguments[i]]) {
			clause.push_back(inst.not_attacked_var[make_pair(ex.id, inst.arguments[i])]);
		}
	}
	clause.push_back(inst.extension_var[ex.id]);
	clauses.push_back(clause);
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (!ex.arg_in_ex[inst.arguments[i]]) {
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (ex.arg_in_ex[inst.arguments[j]]) {
					vector<int> clause;
					clause.push_back(-inst.not_attacked_var[make_pair(ex.id, inst.arguments[i])]);
					clause.push_back(-inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
					clauses.push_back(clause);
				}
			}
		}
	}
	for (int i = 0; i < inst.arguments.size(); i++) {
		if (!ex.arg_in_ex[inst.arguments[i]]) {
			vector<int> clause;
			for (int j = 0; j < inst.arguments.size(); j++) {
				if (ex.arg_in_ex[inst.arguments[j]]) {
					clause.push_back(inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
				}
			}
			clause.push_back(inst.not_attacked_var[make_pair(ex.id, inst.arguments[i])]);
			clauses.push_back(clause);
		}
	}
	}
	return clauses;	
}

vector<vector<int>> prf_neg_clauses(Instance & inst, Example & ex)
{
	vector<vector<int>> clauses = adm_clauses(inst, ex);
	// conflict-free
	for (int i = 0; i < inst.arguments.size(); i++) {
        for (int j = 0; j < inst.arguments.size(); j++) {
            if (!ex.arg_in_ex[inst.arguments[i]] && !ex.arg_in_ex[inst.arguments[j]]) {
                vector<int> clause;
                clause.push_back(-inst.exists_superset_var[ex.id]);
				clause.push_back(-inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
                clause.push_back(-inst.arg_var[make_pair(ex.id,inst.arguments[i])]);
                if (i != j) clause.push_back(-inst.arg_var[make_pair(ex.id,inst.arguments[j])]);
                clauses.push_back(clause);
            } else if (!ex.arg_in_ex[inst.arguments[i]]) {
                vector<int> clause;
                clause.push_back(-inst.exists_superset_var[ex.id]);
				clause.push_back(-inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
                clause.push_back(-inst.arg_var[make_pair(ex.id,inst.arguments[i])]);
                clauses.push_back(clause);
            } else if (!ex.arg_in_ex[inst.arguments[j]]) {
                vector<int> clause;
                clause.push_back(-inst.exists_superset_var[ex.id]);
				clause.push_back(-inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
                clause.push_back(-inst.arg_var[make_pair(ex.id,inst.arguments[j])]);
                clauses.push_back(clause);
            }
        }
    }
    // admissible
    for (int i = 0; i < inst.arguments.size(); i++) {
        if (!ex.arg_in_ex[inst.arguments[i]]) {
            for (int j = 0; j < inst.arguments.size(); j++) {
                if (!ex.arg_in_ex[inst.arguments[j]]) {
                    vector<int> clause;
                    clause.push_back(-inst.exists_superset_var[ex.id]);
					clause.push_back(-inst.attacked_var[make_pair(ex.id,make_pair(inst.arguments[i], inst.arguments[j]))]);
                    for (int k = 0; k < inst.arguments.size(); k++) {
                        if (!ex.arg_in_ex[inst.arguments[k]]) {
                            clause.push_back(inst.attacks_var[make_pair(ex.id,make_pair(inst.arguments[k], inst.arguments[j]))]);
                        } else {
                            clause.push_back(inst.attack_var[make_pair(inst.arguments[k], inst.arguments[j])]);
                        }
                    }
                    clauses.push_back(clause);
                }
            }
        } else {
            for (int j = 0; j < inst.arguments.size(); j++) {
                if (!ex.arg_in_ex[inst.arguments[j]]) {
                    vector<int> clause;
                    clause.push_back(-inst.exists_superset_var[ex.id]);
					clause.push_back(-inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
                    for (int k = 0; k < inst.arguments.size(); k++) {
                        if (!ex.arg_in_ex[inst.arguments[k]]) {
                            clause.push_back(inst.attacks_var[make_pair(ex.id,make_pair(inst.arguments[k], inst.arguments[j]))]);
                        } else {
                            clause.push_back(inst.attack_var[make_pair(inst.arguments[k], inst.arguments[j])]);
                        }
                    }
                    clauses.push_back(clause);
                }
            }
        }
    }
    for (int i = 0; i < inst.arguments.size(); i++) {
        for (int j = 0; j < inst.arguments.size(); j++) {
            if (!ex.arg_in_ex[inst.arguments[i]] && !ex.arg_in_ex[inst.arguments[j]]) {
                vector<int> clause;
                clause.push_back(inst.arg_var[make_pair(ex.id,inst.arguments[i])]);
                clause.push_back(-inst.attacked_var[make_pair(ex.id,make_pair(inst.arguments[i], inst.arguments[j]))]);
                clauses.push_back(clause);
            }
        }
    }
    for (int i = 0; i < inst.arguments.size(); i++) {
        for (int j = 0; j < inst.arguments.size(); j++) {
            if (!ex.arg_in_ex[inst.arguments[i]] && !ex.arg_in_ex[inst.arguments[j]]) {
                vector<int> clause;
                clause.push_back(inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
                clause.push_back(-inst.attacked_var[make_pair(ex.id,make_pair(inst.arguments[i], inst.arguments[j]))]);
                clauses.push_back(clause);
            }
        }
    }
    for (int i = 0; i < inst.arguments.size(); i++) {
        for (int j = 0; j < inst.arguments.size(); j++) {
            if (!ex.arg_in_ex[inst.arguments[i]] && !ex.arg_in_ex[inst.arguments[j]]) {
                vector<int> clause;
                clause.push_back(-inst.arg_var[make_pair(ex.id,inst.arguments[i])]);
                clause.push_back(-inst.attack_var[make_pair(inst.arguments[j], inst.arguments[i])]);
                clause.push_back(inst.attacked_var[make_pair(ex.id,make_pair(inst.arguments[i], inst.arguments[j]))]);
                clauses.push_back(clause);
            }
        }
    }
    for (int i = 0; i < inst.arguments.size(); i++) {
        for (int j = 0; j < inst.arguments.size(); j++) {
            if (!ex.arg_in_ex[inst.arguments[i]] && !ex.arg_in_ex[inst.arguments[j]]) {
                vector<int> clause;
                clause.push_back(inst.arg_var[make_pair(ex.id,inst.arguments[i])]);
                clause.push_back(-inst.attacks_var[make_pair(ex.id,make_pair(inst.arguments[i], inst.arguments[j]))]);
                clauses.push_back(clause);
            }
        }
    }
    for (int i = 0; i < inst.arguments.size(); i++) {
        for (int j = 0; j < inst.arguments.size(); j++) {
            if (!ex.arg_in_ex[inst.arguments[i]] && !ex.arg_in_ex[inst.arguments[j]]) {
                vector<int> clause;
                clause.push_back(inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
                clause.push_back(-inst.attacks_var[make_pair(ex.id,make_pair(inst.arguments[i], inst.arguments[j]))]);
                clauses.push_back(clause);
            }
        }
    }
    for (int i = 0; i < inst.arguments.size(); i++) {
        for (int j = 0; j < inst.arguments.size(); j++) {
            if (!ex.arg_in_ex[inst.arguments[i]] && !ex.arg_in_ex[inst.arguments[j]]) {
                vector<int> clause;
                clause.push_back(-inst.arg_var[make_pair(ex.id,inst.arguments[i])]);
                clause.push_back(-inst.attack_var[make_pair(inst.arguments[i], inst.arguments[j])]);
                clause.push_back(inst.attacks_var[make_pair(ex.id,make_pair(inst.arguments[i], inst.arguments[j]))]);
                clauses.push_back(clause);
            }
        }
    }
    // superset
    vector<int> clause;
    clause.push_back(-inst.exists_superset_var[ex.id]);
    for (int i = 0; i < inst.arguments.size(); i++) {
    	if (!ex.arg_in_ex[inst.arguments[i]]) {
    		clause.push_back(inst.arg_var[make_pair(ex.id,inst.arguments[i])]);
    	}
    }
    clauses.push_back(clause);
	return clauses;
}

vector<vector<int>> pref_additional_clauses(Instance & inst)
{
	vector<vector<int>> clauses;
	for (int i = 0; i < inst.examples.size(); i++) { if (inst.examples[i].pos) {
		for (int j = 0; j < inst.examples.size(); j++) { if (inst.examples[j].pos) {
			if (i != j) {
				bool is_subset = true;
				for (int k = 0; k < inst.examples[i].args.size(); k++) {
					if (!inst.examples[j].arg_in_ex[inst.examples[i].args[k]]) {
						is_subset = false; break;
					}
				}
				if (is_subset && inst.examples[i].args.size() != inst.examples[j].args.size()) {
					vector<int> clause;
					clause.push_back(-inst.extension_var[inst.examples[i].id]);
					clause.push_back(-inst.extension_var[inst.examples[j].id]);
					clauses.push_back(clause);
				}
			}
		} }
	} }
	return clauses;	
}

/*!
 * Wcnf file output.
 *
 * nbvar - number of variables
 * nbclauses - number of clauses
 * top - weight assigned to hard clauses
 * hard_clauses - hard clauses in MaxSAT instance
 * soft_clauses - soft clauses in MaxSAT instance
 * weights - weights of soft clauses
 * filename - name of output file
 */
bool write_wcnf_to_file(int nbvar, int nbclauses, int top, 
                   vector<vector<int>>& hard_clauses, 
                   vector<vector<int>>& soft_clauses,
                   vector<int> & weights, 
                   string filename)
{
    ofstream file(filename);
    if (file.is_open()) {
        file << "p wcnf " << nbvar << " " << nbclauses << " " << top << "\n";
        for (int i = 0; i < hard_clauses.size(); i++) {
            file << top << " ";
            for (int j = 0; j < hard_clauses[i].size(); j++) {
                file << hard_clauses[i][j] << " ";
            }
            file << "0\n";
        }
        for (int i = 0; i < soft_clauses.size(); i++) {
            file << weights[i] << " ";
            for (int j = 0; j < soft_clauses[i].size(); j++) {
                file << soft_clauses[i][j] << " ";
            }
            file << "0\n";
        }
        return true;
    }
    return false;
}

bool write_lp_to_file(int nbvar,
                      vector<vector<int>> & hard_clauses,
                      vector<vector<int>> & soft_clauses,
                      vector<int> & weights,
                      string filename)
{
    ofstream file(filename);
    if (file.is_open()) {
        file << "Minimize\n obj: ";
        for (int i = 1; i <= soft_clauses.size(); i++) {
                if (weights[i-1] == 1) file << "b" << i;
                else file << weights[i-1] << " b" << i;
                if (i != soft_clauses.size()) file << " + ";
        }
        file << "\nSubject To\n";
        for (int i = 0; i < soft_clauses.size(); i++) {
            int count_neg = 0;
            for (int j = 0; j < soft_clauses[i].size(); j++) {
                file << " ";
                if (soft_clauses[i][j] > 0) {
                    if (j != 0) file << " + ";
                    file << "x" << abs(soft_clauses[i][j]);
                } else {
                    if (j != 0) file << " - ";
                    else file << "-";
                    file << "x" << abs(soft_clauses[i][j]);
                    count_neg--;
                }
            }
            file << " + b" << i+1 << " >= " << ++count_neg << "\n";
        }
        for (int i = 0; i < hard_clauses.size(); i++) {
            int count_neg = 0;
            for (int j = 0; j < hard_clauses[i].size(); j++) {
                file << " ";
                if (hard_clauses[i][j] > 0) {
                    if (j != 0) file << "+ ";
                    file << "x" << abs(hard_clauses[i][j]);
                } else {
                    file << "- x" << abs(hard_clauses[i][j]);
                    count_neg--;
                }
            }
            file << " >= " << ++count_neg << "\n";
        }
        file << "Bounds\n";
        for (int i = 1; i <= soft_clauses.size(); i++) {
            file << " 0 <= b" << i << " <= 1\n";
        }
        for (int i = 1; i <= nbvar; i++) {
            file << " 0 <= x" << i << " <= 1\n";
        }
        file << "Binary\n";
        for (int i = 1; i <= soft_clauses.size(); i++) {
            file << " b" << i << "\n";
        }
        for (int i = 1; i <= nbvar; i++) {
            file << " x" << i << "\n";
        }
        file << "End\n";
        return true;
    }
    return false;
}

void synthesize(Instance & inst, string outfile, string type, bool adm_in_abstraction)
{
	vector<vector<int>> hard_clauses;
	vector<vector<int>> soft_clauses;
	vector<int> weights;
	int top = 1;
	bool second_level = false;
	for (int i = 0; i < inst.examples.size(); i++) {
		vector<vector<int>> tmp;
		if (inst.examples[i].sem == "cf") {
			tmp = cf_clauses(inst, inst.examples[i]);
		} else if (inst.examples[i].sem == "adm") {
			tmp = adm_clauses(inst, inst.examples[i]);
		} else if (inst.examples[i].sem == "com") {
			tmp = com_clauses(inst, inst.examples[i]);
		} else if (inst.examples[i].sem == "stb") {
			tmp = stb_clauses(inst, inst.examples[i]);
		} else if (inst.examples[i].sem == "prf") {
			if (inst.examples[i].pos) {
				if (adm_in_abstraction) {
					tmp = adm_clauses(inst, inst.examples[i]);
				} else {
					tmp = com_clauses(inst, inst.examples[i]);
				}
			} else {
				tmp = prf_neg_clauses(inst, inst.examples[i]);
			}
			second_level = true;
		}
		hard_clauses.insert(hard_clauses.end(), tmp.begin(), tmp.end());
		vector<int> clause;
		if (inst.examples[i].pos) {
			clause.push_back(inst.extension_var[inst.examples[i].id]);
		} else {
			if (inst.examples[i].sem != "prf") {
				clause.push_back(-inst.extension_var[inst.examples[i].id]);
			} else {
				clause.push_back(-inst.extension_var[inst.examples[i].id]);
				clause.push_back(inst.exists_superset_var[inst.examples[i].id]);
			}
		}
		soft_clauses.push_back(clause);
		weights.push_back(inst.examples[i].weight);
		top += inst.examples[i].weight;
	}
	// if file specified output wcnf/lp and exit
    if (outfile != "") {
    	if (type == "lp") {
                write_lp_to_file(inst.count, hard_clauses, soft_clauses, weights, outfile);
                return;
        }
        write_wcnf_to_file(inst.count, hard_clauses.size() + soft_clauses.size(),
        				   top, hard_clauses, soft_clauses, weights, outfile);
        return;
    }
    
    // initialize MaxSAT solver
    MaxSAT_Solver maxsat_solver = MaxSAT_Solver();

    // add generated clauses to MaxSAT solver    
    for (int i = 0; i < hard_clauses.size(); i++) {
        maxsat_solver.add_hard_clause(hard_clauses[i]);
    }

    if (!second_level) {
    	for (int i = 0; i < soft_clauses.size(); i++) {
        	maxsat_solver.add_soft_clause(weights[i], soft_clauses[i]);
    	}
	    // generate attack structure from truth assignment
    	maxsat_solver.solve();
    	for (int i = 0; i <= maxsat_solver.assignment.rbegin()->first; i++) {
    		if (inst.var_extension.find(i+1) != inst.var_extension.end()) {
    			inst.examples[inst.example_id_to_index[inst.var_extension[i+1]]].sat = maxsat_solver.assignment[i];
    		}
        	if (inst.var_attack.find(i+1) != inst.var_attack.end() && maxsat_solver.assignment[i]) {
        		inst.attacks.push_back(make_pair(inst.var_attack[i+1].first, inst.var_attack[i+1].second));
        	}
    	}
   	} else {
   		// enter second-level procedure
   		int count = 0;
   		vector<vector<int>> additional_clauses = pref_additional_clauses(inst);
   		for (int i = 0; i < additional_clauses.size(); i++) {
        	maxsat_solver.add_hard_clause(additional_clauses[i]);
    	}
    	for (int i = 0; i < soft_clauses.size(); i++) {
        	maxsat_solver.add_soft_clause(weights[i], soft_clauses[i]);
    	}
    	while (true) {
   			maxsat_solver.solve();
   			count++;
   			AF temp;
   			for (int i = 0; i < inst.arguments.size(); i++) {
   				temp.addArgument(inst.index_to_arg[inst.arguments[i]]);
   			}
   			for (int i = 0; i <= maxsat_solver.assignment.rbegin()->first; i++) {
   				if (inst.var_extension.find(i+1) != inst.var_extension.end()) {
    				inst.examples[inst.example_id_to_index[inst.var_extension[i+1]]].sat = maxsat_solver.assignment[i];
				}
				if (inst.var_attack.find(i+1) != inst.var_attack.end() && maxsat_solver.assignment[i]) {
        			temp.addAttack(make_pair(inst.index_to_arg[inst.var_attack[i+1].first], inst.index_to_arg[inst.var_attack[i+1].second]));
        		}
        		if (inst.var_exists_superset.find(i+1) != inst.var_exists_superset.end() && maxsat_solver.assignment[i]) {
        			inst.examples[inst.example_id_to_index[inst.var_exists_superset[i+1]]].sat = false;
        		}
    		}
    		temp.initialize();
    		bool preferred_ok = true;
    		int example_id = -1;
   			for (int i = 0; i < inst.examples.size(); i++) {
   				// loop through preferred examples that are admissible
   				if (inst.examples[i].sem == "prf" && inst.examples[i].pos && inst.examples[i].sat) {
   					// initialize SAT solver to check if example is preferred
            		SAT_Solver sat_solver = SAT_Solver();
            		vector<vector<int>> check_clauses = Verification::complete_clauses(temp);
            		for (int j = 0; j < inst.arguments.size(); j++) {
                    	if (inst.examples[i].arg_in_ex[inst.arguments[j]]) {
                        	vector<int> clause;
                        	clause.push_back(inst.arguments[j]);
                        	check_clauses.push_back(clause);
                    	}
                	}
                	vector<int> clause;
                	for (int j = 0; j < inst.arguments.size(); j++) {
                    	if (!inst.examples[i].arg_in_ex[inst.arguments[j]]) {
                        	clause.push_back(inst.arguments[j]);
                    	}
                	}
                	check_clauses.push_back(clause);
                	// add generated clauses to SAT solver
                	for (int j = 0; j < check_clauses.size(); j++) {
            			sat_solver.add_clause(check_clauses[j]);
            		}
            		bool sat = sat_solver.solve();
                	inst.examples[i].sat = !sat;
                	// pos example is not preferred
                	if (sat && inst.examples[i].pos) {
                		preferred_ok = false;
                		example_id = inst.examples[i].id;
                		break;
                	}
            	}
   			}
   			if (preferred_ok) {
   				cout << "Number of iterations: " << count << "\n";
   				for (int i = 0; i <= maxsat_solver.assignment.rbegin()->first; i++) {
        			if (inst.var_attack.find(i+1) != inst.var_attack.end() && maxsat_solver.assignment[i]) {
        				inst.attacks.push_back(make_pair(inst.var_attack[i+1].first, inst.var_attack[i+1].second));
        			}
    			}
				break;
   			} else {
   				vector<int> refine_clause;
   				if (inst.examples[example_id].pos) {
   					refine_clause.push_back(-inst.extension_var[example_id]);
   				}
   				for (int i = 0; i <= maxsat_solver.assignment.rbegin()->first; i++) {
   					if (inst.var_attack.find(i+1) != inst.var_attack.end()) {
        				if (maxsat_solver.assignment[i]) {
                            refine_clause.push_back(-(i+1));
                        } else {
                            refine_clause.push_back(i+1);
                        }
        			}
                }
                maxsat_solver.add_hard_clause(refine_clause);
			}
   		}
   	}
}

}
