/*!
 * Copyright (c) <2018> <Andreas Niskanen, University of Helsinki>
 * 
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * 
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "Verification.h"

using namespace std;

namespace Verification {

vector<vector<int>> conflictFree_clauses(AF& af)
{
    vector<vector<int>> clauses;
    for (int i = 0; i < af.atts.size(); i++) {
        vector<int> clause;
        clause.push_back(-af.atts[i].first);
        clause.push_back(-af.atts[i].second);
        clauses.push_back(clause);
    }
    return clauses;
}

vector<vector<int>> admissible_clauses(AF& af)
{
    vector<vector<int>> clauses = conflictFree_clauses(af);
    for (int i = 0; i < af.atts.size(); i++) {
        vector<int> clause;
        clause.push_back(-af.atts[i].second);
        clause.insert(clause.end(), af.attackers[af.atts[i].first].begin(), af.attackers[af.atts[i].first].end());
        clauses.push_back(clause);
    }
    return clauses;
}

vector<vector<int>> complete_clauses(AF& af)
{
    vector<vector<int>> clauses = admissible_clauses(af);
    for (int i = 0; i < af.args.size(); i++) {
        vector<int> clause;
        for (int j = 0; j < af.attackers[af.args[i]].size(); j++) {
            clause.push_back(-af.defendVar[af.attackers[af.args[i]][j]]);
        }
        clause.push_back(af.args[i]);
        clauses.push_back(clause);
    }
    for (int i = 0; i < af.args.size(); i++) {
        vector<int> clause;
        clause.push_back(-af.defendVar[af.args[i]]);
        clause.insert(clause.end(), af.attackers[af.args[i]].begin(), af.attackers[af.args[i]].end());
        clauses.push_back(clause);
    }
    for (int i = 0; i < af.args.size(); i++) {
        for (int j = 0; j < af.attackers[af.args[i]].size(); j++) {
            vector<int> clause;
            clause.push_back(-af.attackers[af.args[i]][j]);
            clause.push_back(af.defendVar[af.args[i]]);
            clauses.push_back(clause);
        }
    }
    return clauses;
}

}